unit IdStreamNetFCL;

{$I IdCompilerDefines.inc}

interface

uses
  IdException, IdGlobal, IdStreamRandomAccess, IdSys, System.IO;

type
  EIdEndOfStream = class(EIdException);

  TIdStreamNetFCL = class(TIdStreamRandomAccess)
  protected
    FFreeStream: Boolean;
    FFCLStream: Stream;
    //
    function GetPosition: Int64; override;
    function GetSize: Int64; override;
    procedure SetSize(const AValue: Int64); override;
    procedure SetPosition(const AValue: Int64); override;
  public
    constructor Create(
      AStream: Stream;
      AFreeStream: Boolean = False
      ); reintroduce; virtual;
    destructor Destroy;
      override;
    function ReadLn(AMaxLineLength: Integer = -1; AExceptionIfEOF: Boolean = FALSE): string; override;
    procedure Skip(
      ASize: Int64
      ); override;
    function ReadBytes(
      var VBytes: TIdBytes;
      ACount: Integer = -1;
      AOffset: Integer = 0;
      AExceptionOnCountDiffer: Boolean = True
      ): Integer; override;
    function ReadString: string; override;
    procedure Write(
      const AValue: string
      ); overload; override;
    procedure Write(
      const ABytes: TIdBytes;
      ACount: Integer = -1
      ); overload; override;
    property FreeStream: Boolean read FFreeStream write FFreeStream;
    property FCLStream: Stream read FFCLStream;
  end;

  TIdNativeStreamBase = Stream;

implementation

uses
  IdStack, IdResourceStrings;

const
  LBUFMAXSIZE = 2048;

constructor TIdStreamNetFCL.Create(
  AStream: Stream;
  AFreeStream: Boolean = False
  );
begin
  inherited Create;
  FFCLStream := AStream;
  FFreeStream := AFreeStream;
end;

function TIdStreamNetFCL.ReadLn(AMaxLineLength: Integer = -1; AExceptionIfEOF: Boolean = FALSE): String;
//TODO: Continue to optimize this function. Its performance severely impacts
// the coders
var
  LBufSize, LStringLen, LResultLen: LongInt;
  LBuf: TIdBytes;
 // LBuf: packed array [0..LBUFMAXSIZE] of Char;
  LStrmPos, LStrmSize: Integer; //LBytesToRead = stream size - Position
  LCrEncountered: Boolean;

  function FindEOL(const ABuf: TIdBytes; var VLineBufSize: Integer; var VCrEncountered: Boolean): Integer;
  var
    i: Integer;
  begin
    Result := VLineBufSize; //EOL not found => use all
    i := 0;
    while i < VLineBufSize do begin
      case ABuf[i] of
        Ord(LF): begin
            Result :=  i; {string size}
            VCrEncountered := TRUE;
            VLineBufSize := i+1;
            break;
          end;//LF
        Ord(CR): begin
            Result := i; {string size}
            VCrEncountered := TRUE;
            inc(i); //crLF?
            if (i < VLineBufSize) and (ABuf[i] = Ord(LF)) then begin
              VLineBufSize := i+1;
            end else begin
              VLineBufSize := i;
            end;
            break;
          end;
      end;
      Inc(i);
    end;
  end;

begin
  SetLength(LBuf, LBUFMAXSIZE);
  if AMaxLineLength < 0 then begin
    AMaxLineLength := MaxInt;
  end;//if
  LCrEncountered := FALSE;
  Result := '';
  { we store the stream size for the whole routine to prevent
  so do not incur a performance penalty with TStream.Size.  It has
  to use something such as Seek each time the size is obtained}
  {4 seek vs 3 seek}
  LStrmPos := FCLStream.Position;
  LStrmSize:= FCLStream.Length;

  if (LStrmSize - LStrmPos) > 0 then begin
    while (LStrmPos < LStrmSize) and not LCrEncountered do begin
      LBufSize := Min(LStrmSize - LStrmPos, LBUFMAXSIZE);
      ReadTIdBytesFromStream(FCLStream, LBuf, LBufSize);
      LStringLen := FindEOL(LBuf, LBufSize, LCrEncountered);
      Inc(LStrmPos, LBufSize);

      LResultLen := Length(Result);
      if (LResultLen + LStringLen) > AMaxLineLength then begin
        LStringLen := AMaxLineLength - LResultLen;
        LCrEncountered := TRUE;
        Dec(LStrmPos,LBufSize);
        Inc(LStrmPos,LStringLen);
      end; //if
      Result := Result + BytesToString(LBuf, 0, LStringLen);
      //SetLength(Result, LResultLen + LStringLen);
      //Move(LBuf[0], PChar(Result)[LResultLen], LStringLen);
    end;//while
    FCLStream.Position := LStrmPos;
  end else begin
    EIdEndOfStream.IfTrue(AExceptionIfEOF, Sys.Format(RSEndOfStream, [ClassName, LStrmPos]));
  end;
end;

{
  Moves the current stream position by a relative number of bytes.
  >0 moves toward the end of stream.
  <0 moves toward the stream origin.
  0 does nothing.
}
procedure TIdStreamNetFCL.Skip(ASize: Int64);
begin
  if ASize <> 0 then begin
    FCLStream.Seek(ASize, SeekOrigin.Current);
  end;
end;

destructor TIdStreamNetFCL.Destroy;
begin
  if FreeStream then begin
    Sys.FreeAndNil(FFCLStream);
  end;
  inherited;
end;

function TIdStreamNetFCL.GetPosition: Int64;
begin
  Result := FCLStream.Position;
end;

procedure TIdStreamNetFCL.SetPosition(const AValue: Int64);
begin
  FCLStream.Position := AValue;
end;

function TIdStreamNetFCL.GetSize: Int64;
begin
  Result := FCLStream.Length;
end;

procedure TIdStreamNetFCL.Write(const ABytes: TIdBytes; ACount: Integer);
begin
  if ACount = -1 then
  begin
    ACount := Length(ABytes);
  end;
  FCLStream.Write(ABytes, 0, ACount);
end;

procedure TIdStreamNetFCL.Write(const AValue: string);
begin
  Write(ToBytes(AValue));
end;

function TIdStreamNetFCL.ReadBytes(var VBytes: TIdBytes; ACount,
  AOffset: Integer; AExceptionOnCountDiffer: Boolean): Integer;
begin
  if ACount = -1 then
  begin
    ACount := FCLStream.Length - FCLStream.Position;
  end;
  if Length(VBytes) < (AOffset+ACount) then
  begin
    SetLength(VBytes, AOffset+ACount);
  end;
  Result := FCLStream.Read(VBytes, AOffset, ACount);
  EIdException.IfTrue(AExceptionOnCountDiffer and (Result <> ACount), RSStreamNotEnoughBytes);
end;

function TIdStreamNetFCL.ReadString: string;
var
  L: Integer;
  LBytes: TIdBytes;
begin
  L := ReadInteger;
  if L > 0 then begin
    ReadBytes(LBytes, L);
    Result := BytesToString(LBytes);
  end else begin
    Result := '';
  end;
end;

procedure TIdStreamNetFCL.SetSize(const AValue: Int64);
begin
  FCLStream.SetLength(AValue);
end;

end.
